"use strict";

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

// @ts-check
// Import types

// use Polyfill for util.promisify in node versions < v8
var promisify = require("./util.promisify");

var vm = require("vm");
var fs = require("fs");
var _ = require("lodash");
var path = require("path");
var loaderUtils = require("loader-utils");

var _require = require("./lib/html-tags"),
    createHtmlTagObject = _require.createHtmlTagObject,
    htmlTagObjectToString = _require.htmlTagObjectToString;

var childCompiler = require("./lib/compiler.js");
var prettyError = require("./lib/errors.js");
var chunkSorter = require("./lib/chunksorter.js");
var getHtmlWebpackPluginHooks = require("./lib/hooks.js").getHtmlWebpackPluginHooks;

var fsStatAsync = promisify(fs.stat);
var fsReadFileAsync = promisify(fs.readFile);

var HtmlWebpackPlugin = function () {
  function HtmlWebpackPlugin(options) {
    _classCallCheck(this, HtmlWebpackPlugin);

    var userOptions = options || {};

    var defaultOptions = {
      template: "auto",
      templateContent: false,
      templateParameters: templateParametersGenerator,
      filename: "index.html",
      hash: false,
      inject: true,
      compile: true,
      favicon: false,
      publicPath: false,
      minify: "auto",
      cache: true,
      showErrors: true,
      chunks: "all",
      excludeChunks: [],
      chunksSortMode: "auto",
      meta: {},
      base: false,
      title: "Webpack App",
      xhtml: false
    };

    this.options = Object.assign(defaultOptions, userOptions);

    // Default metaOptions if no template is provided
    if (!userOptions.template && this.options.templateContent === false && this.options.meta) {
      var defaultMeta = {
        // From https://developer.mozilla.org/en-US/docs/Mozilla/Mobile/Viewport_meta_tag
        viewport: "width=device-width, initial-scale=1"
      };
      this.options.meta = Object.assign({}, this.options.meta, defaultMeta, userOptions.meta);
    }

    // Instance variables to keep caching information
    // for multiple builds
    this.childCompilerHash = undefined;
    /**
     * @type {string | undefined}
     */
    this.childCompilationOutputName = undefined;
    this.assetJson = undefined;
    this.hash = undefined;
    this.version = HtmlWebpackPlugin.version;
  }

  _createClass(HtmlWebpackPlugin, [{
    key: "apply",
    value: function apply(compiler) {
      var _this = this;

      var self = this;
      var isCompilationCached = false;
      /** @type Promise<string> */
      var compilationPromise = void 0;

      this.options.template = this.getFullTemplatePath(this.options.template, compiler.context);

      // convert absolute filename into relative so that webpack can
      // generate it at correct location
      var filename = this.options.filename;
      if (path.resolve(filename) === path.normalize(filename)) {
        this.options.filename = path.relative(compiler.options.output.path, filename);
      }

      // `contenthash` is introduced in webpack v4.3
      // which conflicts with the plugin's existing `contenthash` method,
      // hence it is renamed to `templatehash` to avoid conflicts
      this.options.filename = this.options.filename.replace(/\[(?:(\w+):)?contenthash(?::([a-z]+\d*))?(?::(\d+))?\]/gi, function (match) {
        return match.replace("contenthash", "templatehash");
      });

      // Check if webpack is running in production mode
      // @see https://github.com/webpack/webpack/blob/3366421f1784c449f415cda5930a8e445086f688/lib/WebpackOptionsDefaulter.js#L12-L14
      var isProductionLikeMode = compiler.options.mode === "production" || !compiler.options.mode;

      var minify = this.options.minify;
      if (minify === true || minify === "auto" && isProductionLikeMode) {
        /** @type { import('html-minifier').Options } */
        this.options.minify = {
          // https://github.com/kangax/html-minifier#options-quick-reference
          collapseWhitespace: true,
          removeComments: true,
          removeRedundantAttributes: true,
          removeScriptTypeAttributes: true,
          removeStyleLinkTypeAttributes: true,
          useShortDoctype: true
        };
      }

      // Clear the cache once a new HtmlWebpackPlugin is added
      childCompiler.clearCache(compiler);

      // Register all HtmlWebpackPlugins instances at the child compiler
      compiler.hooks.thisCompilation.tap("HtmlWebpackPlugin", function (compilation) {
        // Clear the cache if the child compiler is outdated
        if (childCompiler.hasOutDatedTemplateCache(compilation)) {
          childCompiler.clearCache(compiler);
        }
        // Add this instances template to the child compiler
        childCompiler.addTemplateToCompiler(compiler, _this.options.template);
        // Add file dependencies of child compiler to parent compiler
        // to keep them watched even if we get the result from the cache
        compilation.hooks.additionalChunkAssets.tap("HtmlWebpackPlugin", function () {
          var childCompilerDependencies = childCompiler.getFileDependencies(compiler);
          childCompilerDependencies.forEach(function (fileDependency) {
            compilation.compilationDependencies.add(fileDependency);
          });
        });
      });

      compiler.hooks.make.tapAsync("HtmlWebpackPlugin", function (compilation, callback) {
        // Compile the template (queued)
        compilationPromise = childCompiler.compileTemplate(self.options.template, self.options.filename, compilation).catch(function (err) {
          compilation.errors.push(prettyError(err, compiler.context).toString());
          return {
            content: self.options.showErrors ? prettyError(err, compiler.context).toJsonHtml() : "ERROR",
            outputName: self.options.filename,
            hash: ""
          };
        }).then(function (compilationResult) {
          // If the compilation change didnt change the cache is valid
          isCompilationCached = Boolean(compilationResult.hash) && self.childCompilerHash === compilationResult.hash;
          self.childCompilerHash = compilationResult.hash;
          self.childCompilationOutputName = compilationResult.outputName;
          callback();
          return compilationResult.content;
        });
      });

      compiler.hooks.emit.tapAsync("HtmlWebpackPlugin", function (compilation, callback) {
        // Get all entry point names for this html file
        var entryNames = Array.from(compilation.entrypoints.keys());
        var filteredEntryNames = self.filterChunks(entryNames, self.options.chunks, self.options.excludeChunks);
        var sortedEntryNames = self.sortEntryChunks(filteredEntryNames, _this.options.chunksSortMode, compilation);
        var childCompilationOutputName = self.childCompilationOutputName;

        if (childCompilationOutputName === undefined) {
          throw new Error("Did not receive child compilation result");
        }

        // Turn the entry point names into file paths
        var assets = self.htmlWebpackPluginAssets(compilation, childCompilationOutputName, sortedEntryNames);

        // If this is a hot update compilation, move on!
        // This solves a problem where an `index.html` file is generated for hot-update js files
        // It only happens in Webpack 2, where hot updates are emitted separately before the full bundle
        if (self.isHotUpdateCompilation(assets)) {
          return callback();
        }

        // If the template and the assets did not change we don't have to emit the html
        var assetJson = JSON.stringify(self.getAssetFiles(assets));
        if (isCompilationCached && self.options.cache && assetJson === self.assetJson) {
          return callback();
        } else {
          self.assetJson = assetJson;
        }

        // The html-webpack plugin uses a object representation for the html-tags which will be injected
        // to allow altering them more easily
        // Just before they are converted a third-party-plugin author might change the order and content
        var assetsPromise = _this.getFaviconPublicPath(_this.options.favicon, compilation, assets.publicPath).then(function (faviconPath) {
          assets.favicon = faviconPath;
          return getHtmlWebpackPluginHooks(compilation).beforeAssetTagGeneration.promise({
            assets: assets,
            outputName: childCompilationOutputName,
            plugin: self
          });
        });

        // Turn the js and css paths into grouped HtmlTagObjects
        var assetTagGroupsPromise = assetsPromise
        // And allow third-party-plugin authors to reorder and change the assetTags before they are grouped
        .then(function (_ref) {
          var assets = _ref.assets;
          return getHtmlWebpackPluginHooks(compilation).alterAssetTags.promise({
            assetTags: {
              scripts: self.generatedScriptTags(assets.js),
              styles: self.generateStyleTags(assets.css),
              meta: [].concat(_toConsumableArray(self.generateBaseTag(self.options.base)), _toConsumableArray(self.generatedMetaTags(self.options.meta)), _toConsumableArray(self.generateFaviconTags(assets.favicon)))
            },
            outputName: childCompilationOutputName,
            plugin: self
          });
        }).then(function (_ref2) {
          var assetTags = _ref2.assetTags;

          // Inject scripts to body unless it set explictly to head
          var scriptTarget = self.options.inject === "head" ? "head" : "body";
          // Group assets to `head` and `body` tag arrays
          var assetGroups = _this.generateAssetGroups(assetTags, scriptTarget);
          // Allow third-party-plugin authors to reorder and change the assetTags once they are grouped
          return getHtmlWebpackPluginHooks(compilation).alterAssetTagGroups.promise({
            headTags: assetGroups.headTags,
            bodyTags: assetGroups.bodyTags,
            outputName: childCompilationOutputName,
            plugin: self
          });
        });

        // Turn the compiled tempalte into a nodejs function or into a nodejs string
        var templateEvaluationPromise = compilationPromise.then(function (compiledTemplate) {
          // Allow to use a custom function / string instead
          if (self.options.templateContent !== false) {
            return self.options.templateContent;
          }
          // Once everything is compiled evaluate the html factory
          // and replace it with its content
          return self.evaluateCompilationResult(compilation, compiledTemplate);
        });

        var templateExectutionPromise = Promise.all([assetsPromise, assetTagGroupsPromise, templateEvaluationPromise])
        // Execute the template
        .then(function (_ref3) {
          var _ref4 = _slicedToArray(_ref3, 3),
              assetsHookResult = _ref4[0],
              assetTags = _ref4[1],
              compilationResult = _ref4[2];

          return typeof compilationResult !== "function" ? compilationResult : self.executeTemplate(compilationResult, assetsHookResult.assets, {
            headTags: assetTags.headTags,
            bodyTags: assetTags.bodyTags
          }, compilation);
        });

        var injectedHtmlPromise = Promise.all([assetTagGroupsPromise, templateExectutionPromise])
        // Allow plugins to change the html before assets are injected
        .then(function (_ref5) {
          var _ref6 = _slicedToArray(_ref5, 2),
              assetTags = _ref6[0],
              html = _ref6[1];

          var pluginArgs = {
            html: html,
            headTags: assetTags.headTags,
            bodyTags: assetTags.bodyTags,
            plugin: self,
            outputName: childCompilationOutputName
          };
          return getHtmlWebpackPluginHooks(compilation).afterTemplateExecution.promise(pluginArgs);
        }).then(function (_ref7) {
          var html = _ref7.html,
              headTags = _ref7.headTags,
              bodyTags = _ref7.bodyTags;

          return self.postProcessHtml(html, assets, { headTags: headTags, bodyTags: bodyTags });
        });

        var emitHtmlPromise = injectedHtmlPromise
        // Allow plugins to change the html after assets are injected
        .then(function (html) {
          var pluginArgs = {
            html: html,
            plugin: self,
            outputName: childCompilationOutputName
          };
          return getHtmlWebpackPluginHooks(compilation).beforeEmit.promise(pluginArgs).then(function (result) {
            return result.html;
          });
        }).catch(function (err) {
          // In case anything went wrong the promise is resolved
          // with the error message and an error is logged
          compilation.errors.push(prettyError(err, compiler.context).toString());
          // Prevent caching
          self.hash = null;
          return self.options.showErrors ? prettyError(err, compiler.context).toHtml() : "ERROR";
        }).then(function (html) {
          // Allow to use [templatehash] as placeholder for the html-webpack-plugin name
          // See also https://survivejs.com/webpack/optimizing/adding-hashes-to-filenames/
          // From https://github.com/webpack-contrib/extract-text-webpack-plugin/blob/8de6558e33487e7606e7cd7cb2adc2cccafef272/src/index.js#L212-L214
          var finalOutputName = childCompilationOutputName.replace(/\[(?:(\w+):)?templatehash(?::([a-z]+\d*))?(?::(\d+))?\]/gi, function (_, hashType, digestType, maxLength) {
            return loaderUtils.getHashDigest(Buffer.from(html, "utf8"), hashType, digestType, parseInt(maxLength, 10));
          });
          // Add the evaluated html code to the webpack assets
          compilation.assets[finalOutputName] = {
            source: function source() {
              return html;
            },
            size: function size() {
              return html.length;
            }
          };
          return finalOutputName;
        }).then(function (finalOutputName) {
          return getHtmlWebpackPluginHooks(compilation).afterEmit.promise({
            outputName: finalOutputName,
            plugin: self
          }).catch(function (err) {
            console.error(err);
            return null;
          }).then(function () {
            return null;
          });
        });

        // Once all files are added to the webpack compilation
        // let the webpack compiler continue
        emitHtmlPromise.then(function () {
          callback();
        });
      });
    }
  }, {
    key: "evaluateCompilationResult",
    value: function evaluateCompilationResult(compilation, source) {
      if (!source) {
        return Promise.reject(new Error("The child compilation didn't provide a result"));
      }
      // The LibraryTemplatePlugin stores the template result in a local variable.
      // To extract the result during the evaluation this part has to be removed.
      source = source.replace("var HTML_WEBPACK_PLUGIN_RESULT =", "");
      var template = this.options.template.replace(/^.+!/, "").replace(/\?.+$/, "");
      var vmContext = vm.createContext(_.extend({ HTML_WEBPACK_PLUGIN: true, require: require }, global));
      var vmScript = new vm.Script(source, { filename: template });
      // Evaluate code and cast to string
      var newSource = void 0;
      try {
        newSource = vmScript.runInContext(vmContext);
      } catch (e) {
        return Promise.reject(e);
      }
      if ((typeof newSource === "undefined" ? "undefined" : _typeof(newSource)) === "object" && newSource.__esModule && newSource.default) {
        newSource = newSource.default;
      }
      return typeof newSource === "string" || typeof newSource === "function" ? Promise.resolve(newSource) : Promise.reject(new Error('The loader "' + this.options.template + "\" didn't return html."));
    }
  }, {
    key: "getTemplateParameters",
    value: function getTemplateParameters(compilation, assets, assetTags) {
      var _this2 = this;

      var templateParameters = this.options.templateParameters;
      if (templateParameters === false) {
        return Promise.resolve({});
      }
      if (typeof templateParameters === "function") {
        var preparedAssetTags = {
          headTags: this.prepareAssetTagGroupForRendering(assetTags.headTags),
          bodyTags: this.prepareAssetTagGroupForRendering(assetTags.bodyTags)
        };
        return Promise.resolve().then(function () {
          return templateParameters(compilation, assets, preparedAssetTags, _this2.options);
        });
      }
      if ((typeof templateParameters === "undefined" ? "undefined" : _typeof(templateParameters)) === "object") {
        return Promise.resolve(templateParameters);
      }
      throw new Error("templateParameters has to be either a function or an object");
    }
  }, {
    key: "executeTemplate",
    value: function executeTemplate(templateFunction, assets, assetTags, compilation) {
      // Template processing
      var templateParamsPromise = this.getTemplateParameters(compilation, assets, assetTags);
      return templateParamsPromise.then(function (templateParams) {
        try {
          // If html is a promise return the promise
          // If html is a string turn it into a promise
          return templateFunction(templateParams);
        } catch (e) {
          compilation.errors.push(new Error("Template execution failed: " + e));
          return Promise.reject(e);
        }
      });
    }
  }, {
    key: "postProcessHtml",
    value: function postProcessHtml(html, assets, assetTags) {
      if (typeof html !== "string") {
        return Promise.reject(new Error("Expected html to be a string but got " + JSON.stringify(html)));
      }
      var htmlAfterInjection = this.options.inject ? this.injectAssetsIntoHtml(html, assets, assetTags) : html;
      var htmlAfterMinification = _typeof(this.options.minify) === "object" ? require("html-minifier-terser").minify(htmlAfterInjection, this.options.minify) : htmlAfterInjection;
      return Promise.resolve(htmlAfterMinification);
    }

    /*
     * Pushes the content of the given filename to the compilation assets
     * @param {string} filename
     * @param {WebpackCompilation} compilation
     *
     * @returns {string} file basename
     */

  }, {
    key: "addFileToAssets",
    value: function addFileToAssets(filename, compilation) {
      filename = path.resolve(compilation.compiler.context, filename);
      return Promise.all([fsStatAsync(filename), fsReadFileAsync(filename)]).then(function (_ref8) {
        var _ref9 = _slicedToArray(_ref8, 2),
            size = _ref9[0],
            source = _ref9[1];

        return {
          size: size,
          source: source
        };
      }).catch(function () {
        return Promise.reject(new Error("HtmlWebpackPlugin: could not load file " + filename));
      }).then(function (results) {
        var basename = path.basename(filename);
        compilation.fileDependencies.add(filename);
        compilation.assets[basename] = {
          source: function source() {
            return results.source;
          },
          size: function size() {
            return results.size.size;
          }
        };
        return basename;
      });
    }
  }, {
    key: "sortEntryChunks",
    value: function sortEntryChunks(entryNames, sortMode, compilation) {
      // Custom function
      if (typeof sortMode === "function") {
        return entryNames.sort(sortMode);
      }
      // Check if the given sort mode is a valid chunkSorter sort mode
      if (typeof chunkSorter[sortMode] !== "undefined") {
        return chunkSorter[sortMode](entryNames, compilation, this.options);
      }
      throw new Error('"' + sortMode + '" is not a valid chunk sort mode');
    }

    /**
     * Return all chunks from the compilation result which match the exclude and include filters
     * @param {any} chunks
     * @param {string[]|'all'} includedChunks
     * @param {string[]} excludedChunks
     */

  }, {
    key: "filterChunks",
    value: function filterChunks(chunks, includedChunks, excludedChunks) {
      return chunks.filter(function (chunkName) {
        // Skip if the chunks should be filtered and the given chunk was not added explicity
        if (Array.isArray(includedChunks) && includedChunks.indexOf(chunkName) === -1) {
          return false;
        }
        // Skip if the chunks should be filtered and the given chunk was excluded explicity
        if (Array.isArray(excludedChunks) && excludedChunks.indexOf(chunkName) !== -1) {
          return false;
        }
        // Add otherwise
        return true;
      });
    }

    /**
     * Check if the given asset object consists only of hot-update.js files
     *
     * @param {{
        publicPath: string,
        js: Array<string>,
        css: Array<string>,
        manifest?: string,
        favicon?: string
      }} assets
     */

  }, {
    key: "isHotUpdateCompilation",
    value: function isHotUpdateCompilation(assets) {
      return assets.js.length && assets.js.every(function (assetPath) {
        return (/\.hot-update\.js$/.test(assetPath)
        );
      });
    }
  }, {
    key: "htmlWebpackPluginAssets",
    value: function htmlWebpackPluginAssets(compilation, childCompilationOutputName, entryNames) {
      var _this3 = this;

      var compilationHash = compilation.hash;
      /**
       * @type {string} the configured public path to the asset root
       * if a path publicPath is set in the current webpack config use it otherwise
       * fallback to a realtive path
       */
      var webpackPublicPath = compilation.mainTemplate.getPublicPath({
        hash: compilationHash
      });
      var isPublicPathDefined = webpackPublicPath.trim() !== "";
      var publicPath = this.options.publicPath ? this.options.publicPath : isPublicPathDefined ? // If a hard coded public path exists use it
      webpackPublicPath : // If no public path was set get a relative url path
      path.relative(path.resolve(compilation.options.output.path, path.dirname(childCompilationOutputName)), compilation.options.output.path).split(path.sep).join("/");
      if (publicPath.length && publicPath.substr(-1, 1) !== "/") {
        publicPath += "/";
      }

      /**
       * @type {{
          publicPath: string,
          js: Array<string>,
          css: Array<string>,
          manifest?: string,
          favicon?: string
        }}
       */
      var assets = {
        // The public path
        publicPath: publicPath,
        // Will contain all js and mjs files
        js: [],
        // Will contain all css files
        css: [],
        // Will contain the html5 appcache manifest files if it exists
        manifest: Object.keys(compilation.assets).find(function (assetFile) {
          return path.extname(assetFile) === ".appcache";
        }),
        // Favicon
        favicon: undefined
      };

      // Append a hash for cache busting
      if (this.options.hash && assets.manifest) {
        assets.manifest = this.appendHash(assets.manifest, compilationHash);
      }

      // Extract paths to .js, .mjs and .css files from the current compilation
      var entryPointPublicPathMap = {};
      var extensionRegexp = /\.(css|js|mjs)(\?|$)/;
      for (var i = 0; i < entryNames.length; i++) {
        var entryName = entryNames[i];
        var entryPointFiles = compilation.entrypoints.get(entryName).getFiles();
        // Prepend the publicPath and append the hash depending on the
        // webpack.output.publicPath and hashOptions
        // E.g. bundle.js -> /bundle.js?hash
        var entryPointPublicPaths = entryPointFiles.map(function (chunkFile) {
          var entryPointPublicPath = publicPath + _this3.urlencodePath(chunkFile);
          return _this3.options.hash ? _this3.appendHash(entryPointPublicPath, compilationHash) : entryPointPublicPath;
        });

        entryPointPublicPaths.forEach(function (entryPointPublicPath) {
          var extMatch = extensionRegexp.exec(entryPointPublicPath);
          // Skip if the public path is not a .css, .mjs or .js file
          if (!extMatch) {
            return;
          }
          // Skip if this file is already known
          // (e.g. because of common chunk optimizations)
          if (entryPointPublicPathMap[entryPointPublicPath]) {
            return;
          }
          entryPointPublicPathMap[entryPointPublicPath] = true;
          // ext will contain .js or .css, because .mjs recognizes as .js
          var ext = extMatch[1] === "mjs" ? "js" : extMatch[1];
          assets[ext].push(entryPointPublicPath);
        });
      }
      return assets;
    }
  }, {
    key: "getFaviconPublicPath",
    value: function getFaviconPublicPath(faviconFilePath, compilation, publicPath) {
      var _this4 = this;

      if (!faviconFilePath) {
        return Promise.resolve(undefined);
      }
      return this.addFileToAssets(faviconFilePath, compilation).then(function (faviconName) {
        var faviconPath = publicPath + faviconName;
        if (_this4.options.hash) {
          return _this4.appendHash(faviconPath, compilation.hash);
        }
        return faviconPath;
      });
    }
  }, {
    key: "getMetaTags",
    value: function getMetaTags() {
      var metaOptions = this.options.meta;
      if (metaOptions === false) {
        return [];
      }
      // Make tags self-closing in case of xhtml
      // Turn { "viewport" : "width=500, initial-scale=1" } into
      // [{ name:"viewport" content:"width=500, initial-scale=1" }]
      var metaTagAttributeObjects = Object.keys(metaOptions).map(function (metaName) {
        var metaTagContent = metaOptions[metaName];
        return typeof metaTagContent === "string" ? {
          name: metaName,
          content: metaTagContent
        } : metaTagContent;
      }).filter(function (attribute) {
        return attribute !== false;
      });
      // Turn [{ name:"viewport" content:"width=500, initial-scale=1" }] into
      // the html-webpack-plugin tag structure
      return metaTagAttributeObjects.map(function (metaTagAttributes) {
        if (metaTagAttributes === false) {
          throw new Error("Invalid meta tag");
        }
        return {
          tagName: "meta",
          voidTag: true,
          attributes: metaTagAttributes
        };
      });
    }
  }, {
    key: "generatedScriptTags",
    value: function generatedScriptTags(jsAssets) {
      return jsAssets.map(function (scriptAsset) {
        return {
          tagName: "script",
          voidTag: false,
          attributes: {
            src: scriptAsset
          }
        };
      });
    }
  }, {
    key: "generateStyleTags",
    value: function generateStyleTags(cssAssets) {
      return cssAssets.map(function (styleAsset) {
        return {
          tagName: "link",
          voidTag: true,
          attributes: {
            href: styleAsset,
            rel: "stylesheet"
          }
        };
      });
    }
  }, {
    key: "generateBaseTag",
    value: function generateBaseTag(baseOption) {
      if (baseOption === false) {
        return [];
      } else {
        return [{
          tagName: "base",
          voidTag: true,
          attributes: typeof baseOption === "string" ? {
            href: baseOption
          } : baseOption
        }];
      }
    }
  }, {
    key: "generatedMetaTags",
    value: function generatedMetaTags(metaOptions) {
      if (metaOptions === false) {
        return [];
      }
      // Make tags self-closing in case of xhtml
      // Turn { "viewport" : "width=500, initial-scale=1" } into
      // [{ name:"viewport" content:"width=500, initial-scale=1" }]
      var metaTagAttributeObjects = Object.keys(metaOptions).map(function (metaName) {
        var metaTagContent = metaOptions[metaName];
        return typeof metaTagContent === "string" ? {
          name: metaName,
          content: metaTagContent
        } : metaTagContent;
      }).filter(function (attribute) {
        return attribute !== false;
      });
      // Turn [{ name:"viewport" content:"width=500, initial-scale=1" }] into
      // the html-webpack-plugin tag structure
      return metaTagAttributeObjects.map(function (metaTagAttributes) {
        if (metaTagAttributes === false) {
          throw new Error("Invalid meta tag");
        }
        return {
          tagName: "meta",
          voidTag: true,
          attributes: metaTagAttributes
        };
      });
    }
  }, {
    key: "generateFaviconTags",
    value: function generateFaviconTags(faviconPath) {
      if (!faviconPath) {
        return [];
      }
      return [{
        tagName: "link",
        voidTag: true,
        attributes: {
          rel: "shortcut icon",
          href: faviconPath
        }
      }];
    }
  }, {
    key: "generateAssetGroups",
    value: function generateAssetGroups(assetTags, scriptTarget) {
      var result = {
        headTags: [].concat(_toConsumableArray(assetTags.meta), _toConsumableArray(assetTags.styles)),
        bodyTags: []
      };
      // Add script tags to head or body depending on
      // the htmlPluginOptions
      if (scriptTarget === "body") {
        var _result$bodyTags;

        (_result$bodyTags = result.bodyTags).push.apply(_result$bodyTags, _toConsumableArray(assetTags.scripts));
      } else {
        var _result$headTags;

        (_result$headTags = result.headTags).push.apply(_result$headTags, _toConsumableArray(assetTags.scripts));
      }
      return result;
    }
  }, {
    key: "prepareAssetTagGroupForRendering",
    value: function prepareAssetTagGroupForRendering(assetTagGroup) {
      var xhtml = this.options.xhtml;
      var preparedTags = assetTagGroup.map(function (assetTag) {
        var copiedAssetTag = Object.assign({}, assetTag);
        copiedAssetTag.toString = function () {
          return htmlTagObjectToString(this, xhtml);
        };
        return copiedAssetTag;
      });
      preparedTags.toString = function () {
        return this.join("");
      };
      return preparedTags;
    }
  }, {
    key: "injectAssetsIntoHtml",
    value: function injectAssetsIntoHtml(html, assets, assetTags) {
      var _this5 = this;

      var htmlRegExp = /(<html[^>]*>)/i;
      var headRegExp = /(<\/head\s*>)/i;
      var bodyRegExp = /(<\/body\s*>)/i;
      var body = assetTags.bodyTags.map(function (assetTagObject) {
        return htmlTagObjectToString(assetTagObject, _this5.options.xhtml);
      });
      var head = assetTags.headTags.map(function (assetTagObject) {
        return htmlTagObjectToString(assetTagObject, _this5.options.xhtml);
      });

      if (body.length) {
        if (bodyRegExp.test(html)) {
          // Append assets to body element
          html = html.replace(bodyRegExp, function (match) {
            return body.join("") + match;
          });
        } else {
          // Append scripts to the end of the file if no <body> element exists:
          html += body.join("");
        }
      }

      if (head.length) {
        // Create a head tag if none exists
        if (!headRegExp.test(html)) {
          if (!htmlRegExp.test(html)) {
            html = "<head></head>" + html;
          } else {
            html = html.replace(htmlRegExp, function (match) {
              return match + "<head></head>";
            });
          }
        }

        // Append assets to head element
        html = html.replace(headRegExp, function (match) {
          return head.join("") + match;
        });
      }

      // Inject manifest into the opening html tag
      if (assets.manifest) {
        html = html.replace(/(<html[^>]*)(>)/i, function (match, start, end) {
          // Append the manifest only if no manifest was specified
          if (/\smanifest\s*=/.test(match)) {
            return match;
          }
          return start + ' manifest="' + assets.manifest + '"' + end;
        });
      }
      return html;
    }

    /**
     * Appends a cache busting hash to the query string of the url
     * E.g. http://localhost:8080/ -> http://localhost:8080/?50c9096ba6183fd728eeb065a26ec175
     * @param {string} url
     * @param {string} hash
     */

  }, {
    key: "appendHash",
    value: function appendHash(url, hash) {
      if (!url) {
        return url;
      }
      return url + (url.indexOf("?") === -1 ? "?" : "&") + hash;
    }

    /**
     * Encode each path component using `encodeURIComponent` as files can contain characters
     * which needs special encoding in URLs like `+ `.
     *
     * @param {string} filePath
     */

  }, {
    key: "urlencodePath",
    value: function urlencodePath(filePath) {
      return filePath.split("/").map(encodeURIComponent).join("/");
    }

    /**
     * Helper to return the absolute template path with a fallback loader
     * @param {string} template
     * The path to the template e.g. './index.html'
     * @param {string} context
     * The webpack base resolution path for relative paths e.g. process.cwd()
     */

  }, {
    key: "getFullTemplatePath",
    value: function getFullTemplatePath(template, context) {
      if (template === "auto") {
        template = path.resolve(context, "src/index.ejs");
        if (!fs.existsSync(template)) {
          template = path.join(__dirname, "default_index.ejs");
        }
      }
      // If the template doesn't use a loader use the lodash template loader
      if (template.indexOf("!") === -1) {
        template = require.resolve("./lib/loader.js") + "!" + path.resolve(context, template);
      }
      // Resolve template path
      return template.replace(/([!])([^/\\][^!?]+|[^/\\!?])($|\?[^!?\n]+$)/, function (match, prefix, filepath, postfix) {
        return prefix + path.resolve(filepath) + postfix;
      });
    }

    /**
     * Helper to return a sorted unique array of all asset files out of the
     * asset object
     */

  }, {
    key: "getAssetFiles",
    value: function getAssetFiles(assets) {
      var files = _.uniq(Object.keys(assets).filter(function (assetType) {
        return assetType !== "chunks" && assets[assetType];
      }).reduce(function (files, assetType) {
        return files.concat(assets[assetType]);
      }, []));
      files.sort();
      return files;
    }
  }]);

  return HtmlWebpackPlugin;
}();

function templateParametersGenerator(compilation, assets, assetTags, options) {
  return {
    compilation: compilation,
    webpackConfig: compilation.options,
    htmlWebpackPlugin: {
      tags: assetTags,
      files: assets,
      options: options
    }
  };
}

// Statics:
/**
 * The major version number of this plugin
 */
HtmlWebpackPlugin.version = 4;

/**
 * A static helper to get the hooks for this plugin
 *
 * Usage: HtmlWebpackPlugin.getHooks(compilation).HOOK_NAME.tapAsync('YourPluginName', () => { ... });
 */
HtmlWebpackPlugin.getHooks = getHtmlWebpackPluginHooks;
HtmlWebpackPlugin.createHtmlTagObject = createHtmlTagObject;

module.exports = HtmlWebpackPlugin;